// GBList.h;

#ifndef INCLUDED_GBLIST_H
#define INCLUDED_GBLIST_H

#include "ListChoice.hpp"
#ifdef USE_OLD_GBLIST
#include "RecordHere.hpp"
#include "DLList.hpp"
#ifndef INCLUDED_DLLIST_H
#include "DLList.hpp"
#endif
#ifndef INCLUDED_BASEGBLIST_H
#include "BaseGBList.hpp"
#endif
#ifndef INCLUDED_INTPAIR_H
#include "intpair.hpp"
#endif
#ifndef INCLUDED_DEBUG1_H
#include "Debug1.hpp"
#endif

template<class T> class const_GBListIterator;
template<class T> class GBListIterator;
template<class T> class GBListRep;

void GBListError();

template<class T>
class GBList {
private:
  friend class GBListRep<T>;
  friend class GBListIterator<T>;
  friend class const_GBListIterator<T>;
  GBListRep<T> * rep;
public:
  typedef GBListIterator<T> iterator;
  typedef const_GBListIterator<T> const_iterator;

// The empty constructor makes a "quick" empty list
  GBList();

// The delayed copy constructor - if "quick" empty list, just make a real copy
  GBList(const GBList<T> & aList);
  ~GBList();
  // Delayed copying
  void operator = (const GBList<T> & aList);

  bool operator == (const GBList<T> & bList) const;
  bool operator != (const GBList<T> & bList) const;

// Constant functions
  int size() const;
  bool empty() const;

  const_GBListIterator<T> begin() const;
  const_GBListIterator<T> end() const;

// psuedo-STL compliant
  intpair Member(const T & t) const;

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 

These functions will cause the delayed copy to execute because the rep is
being modified (at least, it is assumed that it is being modified).

The function DeepCopy() is used to reduce repetitive coding - it checks if
there is more than one instance and duplicates the GBListRep if so.  It
assumes rep is not NULL.

 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

  void DeepCopy();

  void push_back(const T & a);
  void addElement(const T & a, int index);
  void removeElement(int index);
  void pop_front();
  void removeDuplicates();
  void joinTo(const GBList<T> & bList);
  int insertIfNotMember(const T & t);

// Clear is a special case; it doesn't use DeepCopy but checks for >1 instance
  void clear();

// Iterator functions
// For this implementation, a copy is made in order to iterate over a GBList
// with >1 instance
  GBListIterator<T> begin();
  GBListIterator<T> end();
}; /* GBList<T> */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

This is the GBListRep class definition.  This holds the actual pointer to
the DLList and a reference count, and all of the functions.  All fields are
private, since they should only be accessed by the GBList class, a friend.

 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

template <class T>
class GBListRep : public BaseGBList {
private:
  friend class GBList<T>;
  friend class GBListIterator<T>;
  friend class const_GBListIterator<T>;

  GBListRep();
  GBListRep(const GBListRep<T> & aList);
  virtual ~GBListRep();
  void operator = (const GBListRep<T> & aList);
  typedef GBListIterator<T> iterator;
  typedef const_GBListIterator<T> const_iterator;

  bool operator == (const GBListRep<T> & bList) const;
  bool operator != (const GBListRep<T> & bList) const;

  void push_back(const T & a);
  void addElement(const T & a,int index);
  void removeElement(int index);
  void pop_front();
  void removeDuplicates();
  void clear();
  void joinTo(const GBListRep<T> & bList);

  GBListIterator<T> begin();
  GBListIterator<T> end();
  const_GBListIterator<T> begin() const;
  const_GBListIterator<T> end() const;

// psuedo-STL compliant
  int Ifind(const T & a) const;
  int insertIfNotMember(const T & t);

  int refcnt;
  int count() const;
  int increment();
  int decrement();
  DLList<T> * ptrTOaList;
  mutable const_GBListIterator<T> d_helper_1;
  mutable const_GBListIterator<T> d_helper_2;
};

// The delayed copy constructor - if "quick" empty list, just make a real copy
template<class T>
inline GBList<T>::GBList(const GBList<T> & aList) : rep(aList.rep) {
  if(aList.rep) rep->increment();
}

template<class T>
inline bool GBList<T>::operator != (const GBList<T> & bList) const {
  return ! operator==(bList);
}

// Delayed copying
template<class T>
inline void GBList<T>::operator = (const GBList<T> & aList) {
  if (rep != aList.rep) {
    if (rep != 0 && rep->decrement() == 0) {
      RECORDHERE(delete rep;)
    };
    rep = aList.rep;
    if(rep) rep->increment();
  }
}

// Constant functions
template<class T>
inline bool GBList<T>::empty() const {
  return rep ? rep->empty() : true;
}

template<class T>
inline intpair GBList<T>::Member(const T & t) const {
  int n = (rep==0) ? -1 : rep->Ifind(t);
  return intpair(n!=-1, n);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

These functions will cause the delayed copy to execute because the rep is
being modified (at least, it is assumed that it is being modified).

The function DeepCopy() is used to reduce repetitive coding - it checks if
there is more than one instance and duplicates the GBListRep if so.  It
assumes rep is not 0.

 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

template<class T>
inline void GBList<T>::removeElement(int index) {
  if (rep == 0) GBListError();
  if(rep->count()!=1) DeepCopy();
  rep->removeElement(index);
};

template<class T>
inline void GBList<T>::pop_front() {
  if (rep == 0) DBG();
  if(rep->count()!=1) DeepCopy();
  rep->pop_front();
};

template<class T>
inline void GBList<T>::removeDuplicates() {
  if(rep) {
    if(rep->count()!=1) DeepCopy();
    rep->removeDuplicates();
  }
}

template<class T>
inline void GBList<T>::joinTo(const GBList<T> & bList) {
  if(bList.rep) {
    if (!rep) {
      RECORDHERE(rep = new GBListRep<T>();)
      if(!rep) DBG();
    }
    if(rep->count()!=1) DeepCopy();
    if (!bList.rep) rep->joinTo(*(bList.rep));
  };
}

template<class T>
inline int GBList<T>::insertIfNotMember(const T & t) {
  if (rep == 0) {
    RECORDHERE(rep = new GBListRep<T>();)
    if (rep == 0) DBG();
  }
  if(rep->count()!=1) DeepCopy();
  return rep->insertIfNotMember(t);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  These are the GBListRep<T> class inline functions.
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

template<class T>
inline GBListRep<T>::GBListRep() : refcnt(1), ptrTOaList(0) {
  RECORDHERE(ptrTOaList=new DLList<T>();)
  if(ptrTOaList == 0) DBG();
  zeroSize();
}

template<class T>
inline bool GBListRep<T>::operator !=(const GBListRep<T> & second) const {
  return (!((*this) == second));
}
 
template<class T>
inline GBListIterator<T> GBListRep<T>::begin() { 
  return GBListIterator<T>(*this); 
};

template<class T>
inline GBListIterator<T> GBListRep<T>::end() { 
  GBListIterator<T>::end_marker dummy; 
  return GBListIterator<T>(*this,dummy); 
};

template<class T>
inline const_GBListIterator<T> GBListRep<T>::begin() const { 
  return const_GBListIterator<T>(*this); 
}

template<class T>
inline const_GBListIterator<T> GBListRep<T>::end() const { 
  const_GBListIterator<T>::end_marker dummy; 
  return const_GBListIterator<T>(*this,dummy); 
}

// psuedo-STL compliant
template<class T>
inline int GBListRep<T>::insertIfNotMember(const T & t) { 
  int n = Ifind(t);
  if (n==-1) {
    push_back(t); 
    n = _size;
  }
  return n;
};

template<class T>
inline int GBListRep<T>::count() const { return refcnt; }

template<class T>
inline int
GBListRep<T>::increment() { return ++refcnt; }

template<class T>
inline int GBListRep<T>::decrement() { return --refcnt; }

template<class T>
inline GBList<T>::GBList() : rep(0) { }

template<class T>
inline bool
GBList<T>::operator == (const GBList<T> & bList) const {
  bool result;
  if (rep == bList.rep) {
    result = true;
  } else if (rep == 0 || bList.rep == 0) {
    result = false;
  } else if(rep->_size!= bList.rep->_size) {
    result = false;
  } else {
    result = (*rep) == (*(bList.rep));
  }
  return result;
};

template<class T>
inline int GBList<T>::size() const {
  return rep ? rep->size() : 0;
};

template<class T>
inline const_GBListIterator<T> GBList<T>::begin() const {
  return rep ? ((const GBListRep<T> *)rep)->begin() : const_GBListIterator<T>();
}

template<class T>
inline const_GBListIterator<T> GBList<T>::end() const {
  return rep ? ((const GBListRep<T> *)rep)->end() : const_GBListIterator<T>();
};

#include "BaGBListIterator.hpp"

template<class U>
class GBListIterator : public BaGBListIterator {
private:
  friend class GBList<U>;
  friend class GBListRep<U>;
public:
  GBListIterator();  
    // dummy impelementation
public:
  struct end_marker {};
  // iterator to the beginning of the container
  GBListIterator(GBListRep<U> & ptr);
  GBListIterator(GBListRep<U> & ptr,end_marker);
  // iterator starting at the mth element of the container
  GBListIterator(GBListRep<U> & ptr,int m);
  GBListIterator(const GBListIterator<U> & ptr);
  GBListIterator(DLList<U> & ptr);
  GBListIterator(DLList<U> & ptr,Pix p);
  ~GBListIterator();

  U & operator *() const;
  const GBListIterator<U> & operator=(const GBListIterator<U> & x) {
    assign(x); return * this;
  };
};

template <class U>
inline GBListIterator<U>::GBListIterator() : BaGBListIterator() { }

template<class U>
inline GBListIterator<U>::GBListIterator(GBListRep<U> & ptr) : 
     BaGBListIterator(0,*(ptr.ptrTOaList)) {
  _p = _ptr->first();
};

template<class U>
inline GBListIterator<U>::GBListIterator(GBListRep<U> & ptr,end_marker) : 
     BaGBListIterator(0,*(ptr.ptrTOaList)) {
};

template<class U>
inline GBListIterator<U>::GBListIterator(DLList<U> & ptr) :
   BaGBListIterator(0,ptr) {
  _p = _ptr->first();
};
 
template<class U>
inline GBListIterator<U>::GBListIterator(DLList<U> & ptr,Pix p) :
   BaGBListIterator(p,ptr) {};

template<class U>
inline
GBListIterator<U>::GBListIterator(GBListRep<U> & ptr,int m) : 
   BaGBListIterator(0,*(ptr.ptrTOaList)) {
  _p = _ptr->first();
  for(int i=1;i<m;++i) {
    _ptr->next(_p);
  }
};

template<class U>
inline GBListIterator<U>::GBListIterator(const GBListIterator<U> & ptr) : 
     BaGBListIterator(ptr._p, *(ptr._ptr)) {
};

template<class U>
inline U & GBListIterator<U>::operator *() const { 
  if(_p==0) DBG();
  return ((DLList<U> *)(void *)_ptr)->operator()(_p); 
};

template<class U>
class const_GBListIterator { 
  friend class GBList<U>;
  friend class GBListRep<U>;
  friend class GBListIterator<U>;

  // member data
  Pix _p;
  BaseDLList * _ptr;
public:
  struct end_marker {};
  const_GBListIterator() : _p(0), _ptr(0) {};

    // iterator to the beginning of the container
  const_GBListIterator(const GBListRep<U> & ptr);
  const_GBListIterator(const GBListRep<U> & ptr,end_marker); 
    // iterator starting at the mth element of the container
  const_GBListIterator(const GBListRep<U> & ptr,int m);
  const_GBListIterator(const const_GBListIterator<U> & x) : _p(x._p), _ptr(x._ptr) {};
  void operator=(const const_GBListIterator & x) {
    _p = x._p;
    _ptr = x._ptr;
  };
  ~const_GBListIterator() {};
  const U & operator *() const;
  void operator ++();
  void operator ++(int);
  void advance(int n) {
    for(int i=1;i<=n;++i) ++(*this);
  };
  void operator --();
  void operator --(int);
  bool operator ==(const const_GBListIterator & iter) const { return _p==iter._p;};
  bool operator !=(const const_GBListIterator & iter) const { return _p!=iter._p;};
};
#ifndef INCLUDED_GBLIST_H
#include "GBList.hpp"
#endif

template<class T>
inline  const_GBListIterator<T>::const_GBListIterator(const GBListRep<T> & ptr) : _p(0), _ptr(ptr.ptrTOaList) {
   _p = _ptr->first();
};

template<class T>
inline const_GBListIterator<T>::const_GBListIterator(const GBListRep<T> & ptr,end_marker) : _p(0), 
        _ptr(ptr.ptrTOaList) {};

template<class T>
inline const T & const_GBListIterator<T>::operator *() const {
  if(_p==0) DBG();
  return  (*(DLList<T> *) (void *)_ptr)(_p);
};

template<class T> 
inline void const_GBListIterator<T>::operator ++() { 
  _ptr->next(_p);
};

template<class T> 
inline void const_GBListIterator<T>::operator ++(int) { _ptr->next(_p);};

template<class T> 
inline void const_GBListIterator<T>::operator --() {
  if(_p==0) {
    _p = _ptr->last();
  } else {
    _ptr->prev(_p);
  };
};

template<class T> 
inline  void const_GBListIterator<T>::operator --(int) {
  if(_p==0) {
    _p = _ptr->last();
  } else {
    _ptr->prev(_p);
  };
};
#endif 
#endif 
